#! /usr/bin/env python

import re
import string
import jinja2
import yaml

from ipaddr import get_mixed_type_key, Bytes, \
    IPv4Address, IPv6Address, IPAddress, IPNetwork
from .revzones import reverse_name

from . import log
from . import yaml_loader


class DBClass(type):

    """Metaclass for all objects from the configuration database.

    This metaclass provides a dictionary collecting all instances
    of each concrete database object class.
    """

    def __new__(mcs, name, bases, dict):
        cls = super(DBClass, mcs).__new__(mcs, name, bases, dict)

        # DBObject is derived from object, and serves as an abstract base
        # class: it has no instances dict. Each class directly derived from
        # DBObject has one. Further derived classes share their parent's
        # instances dict (e.g. RevZone shares the Zone instance dict).

        if len(bases) > 0 and bases[0] != object:
            instances_dict = getattr(bases[0], '_instances', {})
            setattr(cls, name.lower() + 's', instances_dict)
            setattr(cls, '_instances', instances_dict)
        return cls


class DBObject(object, metaclass=DBClass):

    """Root (abstract) class for all objects from the configuration
    database.
    """

    def add_attr(self, k, v):
        if k == 'flags' and not isinstance(v, list):
            v = v.split(',')

        if isinstance(self.attrs.get(k), list):
            if isinstance(v, list):
                self.attrs[k].extend(v)
            else:
                self.attrs[k].append(v)
        else:
            self.attrs[k] = v

    def add_flag(self, f):
        self.add_attr('flags', f)

    def has_flag(self, f):
        return f in self.attrs.get('flags', [])

    def instance_key(self):
        '''Unique key to identify instances'''

        return self.name

    def __init__(self, name, **kwargs):

        self.name = name
        if 'attrs' not in self.__dict__:
            self.attrs = {}
        self.attrs['flags'] = []

        for k in list(kwargs.keys()):
            v = kwargs[k]
            if not isinstance(v, list):
                v = [v]
            for vv in v:
                self.add_attr(k, vv)

        if isinstance(name, dict):
            # Singleton
            instance_name = None
        else:
            instance_name = self.instance_key()

        if instance_name in type(self)._instances:
            raise Exception("Duplicate %s: %s" % (type(self), instance_name))

        type(self)._instances[instance_name] = self


class Vars(DBObject):
    pass


class Group(DBObject):

    def __init__(self, name, **kwargs):
        self.hosts = []
        # List of Host objects

        super(Group, self).__init__(name, **kwargs)

    def get_attr(self, key):
        return self.attrs.get(key, '')

    def add_host(self, host):
        self.hosts.append(host)

    def del_host(self, host):
        self.hosts.remove(host)


class OrgUnit(DBObject):

    pass


def strip_view(s):
    """Remove view part from a <domain>_<view> specification.
    """

    i = s.find('_')
    if i == -1:
        i = len(s)
    return s[:i]


class Zone(DBObject):

    def __init__(self, name, **kwargs):
        self.rrs = {}
        self.has_root_cname = 0
        self.attrs = {'mx': []}
        super(Zone, self).__init__(name, **kwargs)

    def add_rr(self, key, type, value):
        m = re.match("(?i)((.*)\.|)" + strip_view(self.name), key)
        if m:
            key = m.group(1)
            if key == "":
                key = "@"
            else:
                # Strip dot
                key = key[:-1]

        if key not in self.rrs:
            self.rrs[key] = []
        new_rr = {'type': type, 'value': value}
        if self.rrs[key].count(new_rr) == 0:
            self.rrs[key].append(new_rr)
        return key

    def add_a(self, key, value):
        if isinstance(value, IPv6Address):
            self.add_rr(key, 'AAAA', value)
        else:
            self.add_rr(key, 'A', value)

    def add_mxes(self, fqdn):
        for mx in self.attrs['mx']:
            self.add_rr(fqdn, 'MX', mx)

    def add_ptr(self, inaddr, fqdn):
        if fqdn[-1] != '.':
            fqdn = fqdn + '.'
        self.add_rr(inaddr, 'PTR', fqdn)

    def add_cname(self, alias, name):
        # Name may be an absolute name, or a name relative
        # to the root of this zone.

        self_suffix = "." + strip_view(self.name) + "."
        if name.endswith(self_suffix):
            name = name[:-len(self_suffix)]
        alias_key = self.add_rr(alias, 'CNAME', name)
        if alias_key == "@":
            self.has_root_cname = 1


class RevZone(Zone):

    def __init__(self, name, **kwargs):
        self.prefix = IPNetwork(name)
        super(RevZone, self).__init__(reverse_name(self.prefix), **kwargs)


class LDAPTemplate(DBObject):

    @staticmethod
    def ipvfilter(a, v):
        if isinstance(a, list):
            return [_f for _f in map(LDAPTemplate.ipvfilter, a, [v] * len(a)) if _f]

        elif a.version == v:
            return str(a)

    def __init__(self, name, **kwargs):
        self._tmpl = {}
        self._vars = {}
        self.env = jinja2.Environment()
        self.env.filters['ipv'] = LDAPTemplate.ipvfilter
        self.attrs = {'requires': []}
        super(LDAPTemplate, self).__init__(name, **kwargs)

    def add_attr(self, key, value):
        if key == 'template':
            value = self.env.from_string(value)
        super(LDAPTemplate, self).add_attr(key, value)

    def render(self, tmpl_key, **kwargs):
        for required in self.attrs['requires']:
            if not kwargs.get(required):
                log.vv("No LDAP %s entry for %s (missing %s)" %
                       (tmpl_key, kwargs['name'], required))
                return
        return self.attrs['template'].render(**kwargs)


def ldap_render(ldif, obj=None, **kwargs):
    res = []
    obj_vars = dict(Vars.varss[None].attrs)
    if obj is not None:
        obj_vars.update(obj.__dict__)
        obj_vars.update(obj_vars.pop('attrs'))
    obj_vars.update(kwargs)
    if 'template' in kwargs:
        tmpl_key = kwargs.pop('template')
    else:
        tmpl_key = next(k for (k, Cl) in class_map if isinstance(obj, Cl))

    tmpl = LDAPTemplate.ldaptemplates.get(tmpl_key)
    if tmpl is None:
        return False

    res = tmpl.render(tmpl_key, **obj_vars)
    if not res:
        return False
    if res[-1] != '\n':
        res = res + '\n'
    ldif.append(res)
    return True


# Find the zone(s) in which to output name d for the given view.
# If a view is requested, and views are specified for the domain,
# then only that view is returned.
# If no view is specified for the domain, return the (only) zone for the
# domain.
# If no view is requested, and views are specified for the domain,
# returns ALL views.

def find_zones_for_domain(d, default_view=None, orig_d=None):
    if orig_d is None:
        orig_d = d

    # Strip trailing dot in FQDN

    if d[-1] == ".":
        d = d[:-1]
    d = d.lower()

    # If a view is specified, first try to find a specific zone for the view

    if default_view is not None:
        dv = d + '_' + default_view
        if dv in Zone.zones:
            return [Zone.zones[dv]]

    # No specific view, look for a view-less zone

    if d in Zone.zones:
        return [Zone.zones[d]]

    # No view specified: look for zones with view for domain

    if default_view is None:
        view_prefix = d + '_'
        views = [Zone.zones[z]
                 for z in Zone.zones if z.startswith(view_prefix)]
        if len(views) > 0:
            return views

    # No zone found so far, climb up to parent domain

    dot = d.find(".")
    if dot < 0:
        return []

    return find_zones_for_domain(d[dot + 1:], default_view, orig_d)


def find_reverse_zone(ipa):
    '''Return the reverse zone for ipa'''

    matchlen = 0
    match = None
    for z in list(Zone.zones.values()):
        if isinstance(z, RevZone) \
                and ipa in z.prefix \
                and z.prefix.prefixlen > matchlen:
            matchlen = z.prefix.prefixlen
            match = z
    return match


class IPnet:
    ipnets = []

    def __str__(self):
        return str(self.prefix)

    def __init__(self, prefix):
        self.prefix = IPNetwork(prefix)
        self.hosts_lines = {}
        self.html_hosts_lines = {}
        self._hostmask = int(self.prefix.hostmask)
        if self.prefix.version == 4:
            self.freelist = [self.host_id(h) for h in self.prefix]

    def key(self):
        """Key function for comparisons and sorting
        """

        return get_mixed_type_key(self.prefix)

    def mark_used(self, addr):
        if self.prefix.version == 4:
            try:
                self.freelist.remove(self.host_id(addr))
            except Exception:
                print("Warning: %s already used in %s" % (addr, str(self)))

    def host_id(self, ipa):
        return int(ipa) & int(self._hostmask)

    def set_hosts_line(self, addr, line):
        hid = self.host_id(addr)
        if hid in self.hosts_lines:
            self.hosts_lines[hid] += ' %s' % line
        else:
            self.hosts_lines[hid] = '%s\t%s' % (str(addr), line)

    def set_html_hosts_line(self, addr, line):
        hid = self.host_id(addr)
        if hid not in self.html_hosts_lines:
            self.html_hosts_lines[hid] = ''
        self.html_hosts_lines[hid] += line

    def list_hosts(self):
        if self.prefix.version == 4:
            hosts_list = []
            for h in self.prefix:
                try:
                    hosts_list.append(self.hosts_lines[self.host_id(h)] + '\n')
                except KeyError:
                    hosts_list.append("#%s unused\n" % (h))
            return hosts_list
        else:
            return [l + '\n' for l in list(self.hosts_lines.values())]

    def html_list_hosts(self):
        """ Same as list_hosts, but return an HTML table, and collapse the
            unused entries. """

        if self.prefix.version != 4:
            print("%s is-a %s" % (self.prefix, self.__class__))
            return list(self.html_host_lines.items())

        hosts_list = []
        in_unused_range = False
        unused_first = ""
        unused_last = ""

        for h in self.prefix:
            hid = self.host_id(h)
            line = self.html_hosts_lines.get(hid, None)
            if line is None:
                if in_unused_range:
                    unused_last = h
                else:
                    unused_first = h
                    unused_last = unused_first
                    in_unused_range = True
                continue

            if in_unused_range:
                if unused_first == unused_last:
                    label = str(unused_first)
                else:
                    label = ("%s<br>&nbsp;<small>to</small><br>%s"
                             % (unused_first, unused_last))

                hosts_list.append(
                    '<tr class="unused"><td>' +
                    label +
                    '</td><td>unused</td><td>&nbsp;</td></tr>\n')

            in_unused_range = False
            hosts_list.append(line)

        return hosts_list


class Net(DBObject):

    def __init__(self, name, **kwargs):
        self.dhcp_fixed = []
        self.ipnets = []
        # List of (IPnet, dict) tuples
        self.binat = []
        super(Net, self).__init__(name, **kwargs)

    def prefixes(self, v=None, len=None):
        '''Return prefixes of self.
        If v or len are specified, only return prefixes with
        the corresponding version or length.'''

        return [ipn.prefix for (ipn, _) in self.ipnets
                if (v is None or ipn.prefix.version == v)
                and (len is None or ipn.prefix.prefixlen == len)]

    def add_attr(self, key, value):
        if key == "addr":
            if isinstance(value, dict):
                self.ipnets.extend([(IPnet(k), value[k]) for k in value])
            else:
                self.ipnets.append((IPnet(value), {}))
        elif key == "binat":
            self.binat.append(IPNetwork(value))
        else:
            super(Net, self).add_attr(key, value)


class Host(DBObject):

    def instance_key(self):
        """Provide unique names across entire database

        Two hosts in different organizational units can have the same
        name, so differentiate them using their internal object ids."""

        return "%s.%d" % (self.name, id(self))

    def __init__(self, name, **kwargs):
        self.aliases = []
        self.addrs = []
        self.mac_addrs = []
        self.groups = []
        self.attrs = {'interface': [], 'dhcp': []}

        super(Host, self).__init__(name, **kwargs)

    def add_mac_addr(self, type, value):
        self.mac_addrs.append([type, value])

    def add_alias(self, alias, kind):
        self.aliases.append([alias, kind])

    def add_addr(self, addr):
        a = IPAddress(addr)
        if a not in self.addrs:
            self.addrs.append(a)

    def add_attr(self, key, value):
        if 'decommissioned' in self.attrs:
            return

        if key == "alias":
            self.add_alias(value, "cname")
        elif key == "valias":
            self.add_alias(value, "a")
            self.add_alias(value, "mx")
        elif key == "aalias":
            self.add_alias(value, "a")
        elif key == "mxalias":
            self.add_alias(value, "mx")
        elif key == "addr":
            self.add_addr(value)
        elif key == "ether":
            self.add_mac_addr("ethernet", value)

        elif key == "groups":
            # Parse the list of groups
            r = re.compile("([^,]+),?")
            for g in r.findall(value):
                group = g.strip()
                self.groups.append(group)

                if group in Group.groups:
                    Group.groups[group].add_host(self)
                else:
                    print("Warning, host %s is assigned to unknown group %s" \
                          % (self.name, group))

        elif key == "decommissioned":
            # Remove entry from hosts table, and also from any group that
            # references it.

            self.attrs['decommissioned'] = value
            del Host.hosts[self.instance_key()]
            for g in self.groups:
                if g in Group.groups:
                    Group.groups[g].del_host(self)

        else:
            super(Host, self).add_attr(key, value)

    def get_attr(self, key):
        if key in self.attrs:
            return self.attrs[key]
        else:
            return ""

    def has_address_in(self, prefix):
        for a in self.addrs:
            if a in prefix:
                return True

        return False

    def html(self, addr):
        return """<tr><td>%s</td>
                <td><a href="machines/%s.html">%s</a></td><td>%s</td></tr>
      """ % (addr, self.name, self.name, self.get_attr("purpose") or "&nbsp;")


def find_nets(addr):
    '''Return all nets that match addr and have the longest prefix length'''
    ipa = IPAddress(addr)
    matchlen = 0
    matches = []
    for n in list(Net.nets.values()):
        for ipn, ipnprops in n.ipnets:
            if ipa in ipn.prefix:
                if ipn.prefix.prefixlen > matchlen:
                    matches = []
                    matchlen = ipn.prefix.prefixlen
                if ipn.prefix.prefixlen == matchlen:
                    matches.append((n, ipn, ipnprops))
    return matches


def nat_addr(ipnet, ipa):
    m = ipnet.hostmask
    return IPv4Address(int(ipnet) | (int(ipa) & int(m)))


def find_nat(addr):
    ipa = IPAddress(addr)
    if not isinstance(ipa, IPv4Address):
        return None
    for n in list(Net.nets.values()):
        ipv4_prefixes = n.prefixes(v=4)
        for b in n.binat:
            if ipa in b:
                return nat_addr(ipv4_prefixes[0], ipa)
    return None


def autoconfig_address(p, e):
    if p.version != 6 or p.prefixlen != 64:
        raise Exception("Invalid prefix for SLAAC: %s" % (p,))

    eui48 = [int(b, 16) for b in e.split(':')]
    eui48[0] = eui48[0] & ~3
    eui64 = [eui48[0] | 2] + eui48[1:3] + [0xff, 0xfe] + eui48[3:]
    return IPv6Address(Bytes(p.packed[:8] + b''.join(map(chr, eui64))))


def is_slac(ipa):
    return isinstance(ipa, IPv6Address) and ipa.packed[11:13] == '\xff\xfe'


def get_fqdn(name, addr, nprops):
    default_domain = nprops['domain']
    if addr.version == 6:
        suffix = nprops.get('suffix6', '')
    else:
        suffix = ''

    if name[-1] == ".":
        name_comps = name[:-1].split('.')
        name_comps[0] = name_comps[0] + suffix
        return '.'.join(name_comps)
    else:
        return name + suffix + "." + default_domain

# For a generate block, expand a template of the form:
#   $
# or with optional modifiers:
#   ${offset[,width[,base]]}
#
# Modifiers change the offset from the iterator, field width and base.
# Modifiers are introduced by a { immediately following the $ as
# ${offset[,width[,base]]}. e.g. ${-20,3,d} which subtracts 20 from the current
# value, prints the result as a decimal in a zero padded field of with 3.
# Available output forms are decimal (d), octal (o) and hexadecimal (x or X
# for uppercase). The default modifier is ${0,0,d}.
# (consistent with BIND's $GENERATE directive).


generator_pattern = re.compile('\$(\{([0-9-]*)(,([0-9]*)(,([doxX]))?)?\})?')


def expand_index(m, index):
    offset = 0
    width = 0
    base = 'd'

    if m.group(2):
        offset = int(m.group(2))
    if m.group(4):
        width = int(m.group(4))
    if m.group(6):
        base = m.group(6)

    template = "%%0%d%c" % (width, base)
    return template % (index + offset)


def expand_index_refs(obj, index):
    if isinstance(obj, dict):
        return dict([(k_v[0], expand_index_refs(k_v[1], index)) for k_v in list(obj.items())])

    elif isinstance(obj, list):
        return [expand_index_refs(v, index) for v in obj]

    else:
        return re.sub(generator_pattern,
                      lambda m: expand_index(m, index),
                      obj)


def enter_object(obj):
    db_obj = None
    for k, Cl in class_map:
        if k in obj:
            db_obj = Cl(obj.pop(k), **obj)
            break
    if db_obj is None:
        log.err("Invalid object:\n%s" % obj)


# Note: order is significant:
#   'ldap' object has 'host' and 'group' attributes
#   'host' objects may have a 'net' attribute

class_map = [('ldap', LDAPTemplate),
             ('org-unit', OrgUnit),
             ('host', Host),
             ('net', Net),
             ('group', Group),
             ('zone', Zone),
             ('rev-zone', RevZone),
             ('vars', Vars)]


def load_db(f):
    for obj in yaml.load(f, Loader=yaml_loader.UniqueKeyLoader):
        if 'generate' in obj:
            g = obj['generate']
            index, hi = list(map(int, g.pop('range').split('-')))

            while index <= hi:
                enter_object(expand_index_refs(g, index))
                index = index + 1
        else:
            enter_object(obj)
