#! /usr/bin/env python

# ldapdip
# LDAP Diff and Patch
# $Id: ldapdip.py 253716 2017-01-18 14:47:26Z quinot $

import argparse
import ldap
import ldap.modlist
import ldapurl
import ldif
import os.path
import sys
import yaml

conf = {}
success = True

# Always exclude internal attributes
# (useful when comparing slapcat dumps)

exclude = set(map(str.lower, [
    'structuralObjectClass',
    'entryUUID',
    'creatorsName',
    'createTimestamp',
    'entryCSN',
    'modifiersName',
    'modifyTimestamp',
    'contextCSN'
]))

# If non-empty, consider only these attributes

include = set()


def vv(str):
    """Output str if verbose mode is specified."""
    if conf.get('verbose'):
        print(str, file=sys.stderr)


# Editors provide functions to be called for each operation required to
# make the OLD tree identifcal to the NEW tree.

class LDIFEditor:

    """The LDIF Editor produces an LDIF change record stream."""

    def __init__(self, outf):
        self._ldw = ldif.LDIFWriter(outf)

    def output_entry(self, dn, arg):
        self._ldw.unparse(dn, arg)

    def add_entry(self, dn, modlist):
        self.output_entry(dn, modlist)

    def mod_entry(self, dn, modlist):
        self.output_entry(dn, modlist)

    def del_entry(self, dn):
        self.output_entry(dn, {'changetype': ['delete']})


class LDAPEditor:

    """The LDAP Editor applies the modifications to an LDAP server."""

    def __init__(self, lo):
        self._lo = lo

    def add_entry(self, dn, modlist):
        self._lo.add_s(dn, modlist)

    def mod_entry(self, dn, modlist):
        self._lo.modify_s(dn, modlist)

    def del_entry(self, dn):
        self._lo.delete_s(dn)


def open_ldap(uri, update):
    """Open LDAP server URI. If update is True, also return an LDAPEditor
    based on this connection."""

    vv('Querying LDAP server: %s' % uri)
    lo = ldap.initialize(uri)
    lo.simple_bind_s(conf.get('binddn', ''), conf.get('bindpw', ''))

    args = {}

    lf = conf.get('filter', None)
    if lf is not None:
        args['filterstr'] = lf

    entries = {}

    base = conf.get('base', '')
    if not isinstance(base, list):
        base = [base]

    for b in base:
        lr = lo.search_s(b, ldap.SCOPE_SUBTREE, **args)
        entries.update(dict(lr))

    return entries, (LDAPEditor(lo) if update else None)


class LDIFDictLoader(ldif.LDIFParser):

    """LDIF parser that populates a dict of entries indexed by DN"""

    def __init__(self, *args):
        ldif.LDIFParser.__init__(self, *args)
        self.entries = {}

    def handle(self, dn, entry):
        self.entries[dn] = entry


def open_ldif(uri, update):
    """Open an LDIF file. update must be False (an LDIF file cannot be
    edited), and no editor is ever returned."""

    if update:
        raise Exception('cannot update %s' % uri)
    vv('Loading LDIF file: %s' % uri)
    parser = LDIFDictLoader(open(uri))
    parser.parse()
    return parser.entries, None


def open_tree(tree, update):
    '''Open an LDAP URI or LDIF file, or return a tree object unchanged.'''

    if isinstance(tree, str):
        if ldapurl.isLDAPUrl(tree):
            return open_ldap(tree, update)
        else:
            return open_ldif(tree, update)

    else:
        return tree, None


def handle_exception(dn, op, e):
    """Output an exception message and set global variable success to False"""

    global success

    print("Exception raised trying to %s %s:\n%s" % (op, dn, e), file=sys.stderr)
    success = False


def cleanup(e):
    """Remove from entry e all attributes that are to be ignored"""

    for ea in list(e.keys()):
        eal = ea.lower()
        if (len(include) > 0 and eal not in include) \
                or (eal in exclude):
            e.pop(ea)


def dn_key(dn):
    """Key function used for sorting DNs by rightmost component first."""
    return list(reversed(ldap.dn.explode_dn(dn)))

def diff_entries(old_tree, new_tree, update, out, editor=None):
    """Recursively compare the given trees and compute modification operations
    that must be applied to the old tree to make it identical to the new one.

    If out is not None, send modification operations there;
    if out is None, and update is True, apply modifications
    to old_tree directly.
    If editor is not None, changes are also sent to that editor."""

    old_entries, ed = open_tree(old_tree, update and out is None)
    new_entries, _ = open_tree(new_tree, False)

    if out is not None:
        ed = LDIFEditor(out)

    editors = []
    if ed is not None:
        editors.append(ed)
    if editor is not None:
        if isinstance(editor, list):
            editors.extend(editor)
        else:
            editors.append(editor)

    old_dns = set(old_entries.keys())
    new_dns = set(new_entries.keys())

    # Note: deletions need to be processed in reverse order because you
    # can only delete a leaf object (i.e. one that has no children).

    for dn in sorted(old_dns - new_dns, key=dn_key, reverse=True):
        for ed in editors:
            try:
                ed.del_entry(dn)
            except Exception as e:
                handle_exception(dn, 'delete', e)

    for dn in sorted(old_dns & new_dns, key=dn_key):
        oe = old_entries[dn]
        cleanup(oe)
        ne = new_entries[dn]
        cleanup(ne)

        # Compute differences between oe and ne as a list of
        # LDAP modification operations that transform oe into ne.
        # Note that oe and ne might be identical in LDAP sense
        # (i.e. yield an empty mod list) even if oe != ne because
        # some differences, such as the order of values for
        # multi-valued attributes, are irrelevant for LDAP.

        mod = ldap.modlist.modifyModlist(oe, ne)
        if len(mod) > 0:
            for ed in editors:
                ed.mod_entry(dn, mod)

    for dn in sorted(new_dns - old_dns, key=dn_key):
        for ed in editors:
            try:
                ed.add_entry(dn, ldap.modlist.addModlist(new_entries[dn]))
            except Exception as e:
                handle_exception(dn, 'add', e)


def update_incl_excl(ie, args):
    """Add all attributes in comma-separated list args to set ie"""

    for arg in args:
        ie.update(arg.lower().split(','))


def main():
    global conf

    parser = argparse.ArgumentParser(
        description='Compute differences between two LDAP directories')
    parser.add_argument('old', help='old directory URI')
    parser.add_argument('new', help='new directory URI')

    parser.add_argument('--update', '-u', action='store_true', default=False,
                        help='update to OLD to sync it with NEW')
    parser.add_argument('--verbose', '-v', action='store_true', default=False,
                        help='write verbose messages to stderr')
    parser.add_argument('--output', '-o', action='store', default='-',
                        help='output LDIF file (default stdout)')
    parser.add_argument('--conf', '-c', action='append',
                        default=['ldapdip.conf'],
                        help='configuration file')
    parser.add_argument('--option', '-O', action='append', default=[],
                        help='override configuration option')
    parser.add_argument('--include', '-i', action='append', default=[],
                        help='include only these attributes in comparison')
    parser.add_argument('--exclude', '-x', action='append', default=[],
                        help='exclude these attributes from comparison')

    args = parser.parse_args()

    conf = {}

    # Load default config file (ldapdip.conf), as well as any additional
    # one specified on the command line.

    for conf_file in args.conf:
        if os.path.exists(conf_file):
            conf.update(yaml.load(open(conf_file, 'r').read()))

    # Configuration overrides from command line

    if args.verbose:
        conf['verbose'] = True

    for confopt in args.option:
        optname = confopt[:confopt.index('=')]
        optval = confopt[confopt.index('=') + 1:]
        conf[optname] = optval

    # Inclusion/exclusion processing
    # Should also allow setting these from conf???

    update_incl_excl(include, args.include)
    update_incl_excl(exclude, args.exclude)

    # Prepare output

    if args.update:
        # Updating directory in place, no LDIF output
        out = None

    else:
        if args.output == '-':
            out = sys.stdout
        else:
            out = open(args.output, 'w')

    # Work!

    diff_entries(args.old, args.new, args.update, out)

if __name__ == '__main__':
    main()
    if not success:
        sys.exit(1)
