#! /usr/bin/env python

# hostdb
# Hosts database processing
# $Id: main.py 265777 2019-01-15 11:33:46Z quinot $


from . import db as db
from . import log as log

from .db import OrgUnit, Host, Group, Net, Zone

import argparse
import os
import re
import socket
import sys
import time

from ipaddr import IPv4Address, IPv6Address, IPAddress
from .revzones import reverse_name

#reload(sys)
#sys.setdefaultencoding('utf-8')

autogen_marker = ";;; Automatically generated -- do not edit!\n"
dhcpd_conf_template = 'dhcpd.conf.in'

html_subdir = 'html'
csv_subdir = 'csv'
csv_sep = ','
html_machines_subdir = os.path.join(html_subdir, 'machines')


def html_header(title, subdir=0):
    """Return the html header for a page.

    :param str title: Title of the page
    :param int subdir: Depth of subdirectory containing the page
                       (relative to that containing hosts.css).
    """

    css_path = os.path.join(*([".."] * subdir + ["hosts.css"]))

    return """<html><head><meta charset="UTF-8"><title>%s</title>
  <link href="%s" rel="stylesheet" type="text/css">
  </head><body>""" % (title, css_path)


newserial = None


def get_newserial():
    global newserial
    if newserial is not None:
        return newserial

    today = int(time.strftime("%Y%m%d00", time.localtime(time.time())))
    try:
        serial_file = open("SERIAL", "r+")
        oldserial = int(serial_file.readline())
    except IOError:
        # Case where serial file does not exist yet: create one, and use today
        # as the new serial.
        serial_file = open("SERIAL", "w")
        oldserial = today - 1

    if (today - oldserial) > 0:
        newserial = str(today)
    else:
        newserial = str(oldserial + 1)
    serial_file.seek(0)
    serial_file.truncate()
    serial_file.write(str(newserial + '\n'))
    serial_file.close()

    return newserial


def gen_file(name, lines):
    open(name + '.new', 'w').writelines(lines)
    try:
        os.rename(name, name + '~')
    except OSError:
        pass
    os.rename(name + '.new', name)
    os.chmod(name, 0o444)


def main():
  # 0. Command line

  parser = argparse.ArgumentParser('Process the hosts database')
  parser.add_argument('-f', '--force', action='store_true', default=False,
                      help='force bumping serial number')
  parser.add_argument('-v', '--verbose', action='store_true', default=False,
                      help='verbose output')
  args = parser.parse_args()

  log.verbose = args.verbose

  # 1. Parse input

  db.load_db(sys.stdin)

  # 2. Populate zones and hostfile

  hostfile = [
      "### Automatically generated, do not edit!\n",
      "\n",
      "::1\tlocalhost localhost.my.domain\n",
      "127.0.0.1\tlocalhost localhost.my.domain\n"]

  dhcpfile = ["### Automatically generated, do not edit!\n", "\n"] \
      + open(dhcpd_conf_template, 'r').readlines()

  ldif = ["### Automatically generated, do not edit!\n", "\n"]

  # Generate any required entries for base OUs

  for ou in list(OrgUnit.orgunits.values()):
      db.ldap_render(ldif, ou)

  # Set default domain for networks based on their OU

  for n in list(Net.nets.values()):
      if 'domain' not in n.attrs \
              and 'ou' in n.attrs \
              and 'domain' in OrgUnit.orgunits[n.attrs['ou']].attrs:
          n.add_attr('domain', OrgUnit.orgunits[n.attrs['ou']].attrs['domain'])

  hostfile_html = [html_header("Hosts")]

  # Generate additional hosts for interfaces

  for h in list(Host.hosts.values()):
      for i in h.get_attr('interface'):
          # Prefix name and aliases with host name if they start with an hyphen

          for i_attr in i:
              if (i_attr == 'name' or i_attr.endswith('alias')) \
                      and i[i_attr][0] == '-':
                  i[i_attr] = '%s%s' % (h.name, i[i_attr])

          # Add new host

          Host(i.pop('name'), **i)

  # Group processing pass 1 (before host processing)
  #    * optionally, generate aliases of the form "<groupname>-<hostid>"

  for g in list(Group.groups.values()):
      if g.has_flag("aliases"):
          last = 0
          for h in g.hosts:
              h.add_alias("%s-%d" % (g.name, last), "cname")
              last = last + 1

  # Host processing main loop

  for h in list(Host.hosts.values()):

      no_ldap = h.has_flag('noldap')

      # Add implicit NAT addresses

      real_addrs = h.addrs[:]
      if not h.has_flag("nonat"):
          for a in real_addrs:
              nat_a = db.find_nat(a)
              if nat_a:
                  h.add_attr('addr', nat_a)

      # If no address is specified, try to fetch external ones

      if len(h.addrs) == 0 and h.name[-1] == '.':
          for ai in socket.getaddrinfo(h.name, None):
              # If ai's address family is unknown (e.g. case of Python built
              # without IPv6 support) then ai[4][0] is the address family (int),
              # not the canonical address representation (str).
              if isinstance(ai[4][0], str):
                  h.add_addr(ai[4][0])

      dhcp_addresses = {}
      primary_net = None
      for a in h.addrs:
          # print "%s -> %s" % (h.name, a)

          # Identify what network this address belongs to.
          # This is normally just the unique network with the longest
          # address match. However some IP prefixes appear in more than
          # one network (e.g. in the case of several distinct logical
          # subnets that share the same broadcast domain, and therefore
          # the same IPv6 SLAAC prefix). In this case, the ambiguity
          # can be resolved if another address of the same host uniquely
          # identifies one of these networks ("primary" network).
          # Note that interfaces have already been expanded to distinct
          # Host objects, so the primary network for a base Host is
          # (correctly) not taken into account for its dependent interfaces.

          n, ipnet, ipnprops = None, None, None
          nets = db.find_nets(a)
          ambiguous = (len(nets) > 1)
          for n, ipnet, ipnprops in nets:
              # The primary net is the first one that is unambiguously
              # identified for this host.

              if primary_net is None and len(nets) == 1:
                  primary_net = n

              # If one or more nets were found, and the primary one
              # is among them, return it.

              if primary_net == n:
                  ambiguous = False
                  break

          # If more that one net was found, but none matches the
          # primary one, we have an unresolvable ambiguity.

          if ambiguous:
              n = None

          if n is None:
              try:
                  n = Net.nets[h.attrs['net']]
              except KeyError:
                  log.err("Can't find network for %s (%s)" % (h.name, a))
                  n, ipnet = None, None

          ipa = IPAddress(a)
          if ipnet is not None:
              ipnet.mark_used(ipa)

          host_nicks = []
          if n is None:
              fqdn = h.name

              # If network is unknown, assume any name is for external view

              nprops = {'view': 'ext'}

          else:
              # Set organizational unit from network

              if 'ou' in n.attrs:
                  n_ou = n.attrs['ou']
                  h_ou = h.attrs.setdefault('ou', n_ou)
                  if h_ou != n_ou:
                      print("Host %s OU %s inconsistent with network OU %s" \
                          % (h.name, h_ou, n_ou))

              if h.name[-1] != '.':
                  host_nicks.extend([h.name, '%s.%s' % (h.name, n.name)])

              # Get network attributes, but allow per-IPnet overrides

              nprops = dict(n.attrs)
              if ipnprops is not None:
                  nprops.update(ipnprops)

              fqdn = db.get_fqdn(h.name, ipa, nprops)

              if n.has_flag("dhcp") and not db.is_slac(ipa):
                  if n not in dhcp_addresses:
                      dhcp_addresses[n] = []
                  dhcp_addresses[n].append(a)

              if h.has_flag("64ac") and n.has_flag("64ac"):
                  for p in n.prefixes(v=6, len=64):
                      if isinstance(ipa, IPv4Address):
                          h.addrs.append(IPv6Address(str(p.ip) + str(a)))

              elif n.has_flag('slaac') and not h.has_flag('noslaac'):
                  for p in n.prefixes(v=6, len=64):
                      if not h.has_address_in(p):
                          for (htype, haddr) in h.mac_addrs:
                              if htype == 'ethernet':
                                  h.addrs.append(db.autoconfig_address(p, haddr))

          host_nicks = [fqdn] + host_nicks

          # DNS direct records

          view = nprops.get('view', None)
          if not h.has_flag("nodirect"):
              for dz in db.find_zones_for_domain(fqdn, view):
                  dz.add_a(fqdn, ipa)
                  if h.has_flag("acme-cname") and "acme-cname" in dz.attrs:
                      dz.add_cname("_acme-challenge." + fqdn,
                                   fqdn + "." + dz.attrs["acme-cname"])
                  if not h.has_flag("nomx"):
                      dz.add_mxes(fqdn)

              for [alias, kind] in h.aliases:
                  if n is not None:
                      if alias[-1] != '.':
                          host_nicks.extend([alias, '%s.%s' % (alias, n.name)])
                      alias = db.get_fqdn(alias, ipa, nprops)

                  for alias_dz in db.find_zones_for_domain(alias, view):
                      if kind == "mx":
                          alias_dz.add_mxes(alias)
                      else:
                          # "a" or "cname"

                          host_nicks.append(alias)
                          if kind == "a":
                              alias_dz.add_a(alias, ipa)
                          else:
                              alias_dz.add_cname(alias, fqdn + ".")
                          if h.has_flag("acme-cname"):
                              alias_dz.add_cname("_acme-challenge." + alias,
                                                 alias + dz.attrs["acme-cname"])

          # DNS reverse records

          if (ipnet is not None) and not (h.has_flag('noreverse') or
                                          n.has_flag('noreverse')):
              rz = db.find_reverse_zone(ipnet.prefix)
              if rz is not None:
                  rz.add_ptr(reverse_name(rz.prefix, ipa), fqdn)

          hostline = ''
          for nick in host_nicks:
              if len(hostline) > 0:
                  hostline += ' '
              if nick[-1] == '.':
                  nick = nick[:-1]
              hostline = hostline + re.sub('@\.', '', nick)
          if ipnet is not None:
              ipnet.set_html_hosts_line(ipa, h.html(a))
              ipnet.set_hosts_line(ipa, hostline)
          else:
              hostfile.append('%s\t%s\n' % (ipa, hostline))

      # DHCP

      alias_index = 0
      for (mac_type, mac_value) in h.mac_addrs:
          for (n, d_addresses) in list(dhcp_addresses.items()):
              n.dhcp_fixed.append(
                  'host %s-%u {\n   option host-name "%s";\n   hardware %s %s;\n'
                  % (h.name, alias_index, h.name, mac_type, mac_value))

              # Pending debugging, IPv6 addresses are commented out
              for (cl, prefix, suffix) in \
                      [(IPv4Address, '', ''), (IPv6Address, '#', '6')]:
                  cl_addresses = [str(a)
                                  for a in d_addresses if isinstance(a, cl)]
                  if len(cl_addresses) > 0:
                      n.dhcp_fixed.append(
                          "%s   fixed-address%s %s;\n" %
                          (prefix, suffix, ', '.join(cl_addresses)))

              for o in h.attrs.get('dhcp'):
                  n.dhcp_fixed.append('   ' + o + ';\n')
              n.dhcp_fixed.append('}\n\n')
              alias_index = alias_index + 1

      # Generate LDAP entry

      if not (no_ldap or 'ou' not in h.attrs):
          if db.ldap_render(ldif, h):
              for alias, kind in h.aliases:
                  if kind != 'mx':
                      db.ldap_render(
                          ldif, h, template='alias', alias=alias, kind=kind)
          else:
              # Host did not render: set noldap flag to prevent it from
              # being referenced in a host group.

              # print "%s failed to render for LDAP" % h.name
              h.add_flag('noldap')

  for z in list(Zone.zones.values()):
      if not z.has_root_cname:
          z.add_mxes("@")

  # Group processing pass 2 (after host processing):
  #    * generate LDAP entry
  #    * generate Ansible inventory
  #    * remove hosts from "ungrouped" set

  ungrouped_hosts = [h for h in list(Host.hosts.values())
                     if not h.has_flag('noldap')]
  ansible_inventory = []

  for g in list(Group.groups.values()):
      ansible_inventory.extend(["", "[%s]" % (g.name)])
      ldap_hosts = {}
      # Dict indexed by OU, elements are lists of Host objects

      for h in g.hosts:
          if h in ungrouped_hosts:
              ungrouped_hosts.remove(h)

          if not (h.has_flag('noldap') or 'ou' not in h.attrs):
              ldap_hosts.setdefault(h.attrs['ou'], []).append(h)
              ansible_inventory.append(h.instance_key())

      for g.ou, g.ldap_hosts in list(ldap_hosts.items()):
          db.ldap_render(ldif, g)


  open('ansible_inventory', 'w').write(
      '\n'.join([h.instance_key() for h in ungrouped_hosts] + ansible_inventory))

  # 3. Output files

  open('hostdb.ldif', 'w').write('\n'.join(ldif))
  for z in list(Zone.zones.values()):

      # Zone head file
      hfn = "hd." + z.name

      # Generated zone file
      zfn = "db." + z.name

      # Read zone head, identify serial placeholder

      nserial_seen = 0
      nzone = [autogen_marker, "\n"]
      ln = len(nzone)
      for l in open(hfn, "r").readlines():
          nzone.append(l)
          if re.match("\s+SERIALNO\s+", l):
              nserial_seen = 1
              serial_ln = ln
              serial_placeholder = l
          ln = ln + 1
      if not nserial_seen:
          print("Could not find serial number placeholder in %s!" % (hfn))
          sys.exit(1)

      # Try to load old zone file, and replace serial with placeholder (used to
      # detect if this zone has changed)

      try:
          ozone_file = open(zfn, "r")
          ozone = ozone_file.readlines()
          ozone_file.close()
          for ln in range(0, len(ozone)):
              if re.match("\s+[0-9]+\s+;\s+serial", ozone[ln]):
                  ozone[ln] = serial_placeholder
      except IOError:
          print("No existing zone info for " + zfn)
          ozone = []

      nzone.append("\n")
      nzone.append(";;; Body of zone " + z.name + "\n")

      keys = list(z.rrs.keys())
      keys.sort()
      for k in keys:
          first = 1
          for rr in z.rrs[k]:
              if first:
                  nzone.append('\n')
                  key = k
                  first = 0
              else:
                  key = "\t" * (len(k) / 8)
              nzone.append("%s\tIN\t" % (key) + "%(type)s\t%(value)s\n" % rr)

      if nzone != ozone or args.force:
          # Record forced indication

          forced = " (serial bump forced)" if nzone == ozone else ""

          # Now substitute serial in new zone

          nzone[serial_ln] = re.sub(
              "SERIALNO", get_newserial(), nzone[serial_ln])
          nzone_file = open(zfn + ".new", "w")
          nzone_file.writelines(nzone)
          nzone_file.close()

          try:
              os.rename(zfn, zfn + "~")
          except OSError:
              pass
          os.rename(zfn + ".new", zfn)
          os.chmod(zfn, 0o444)

          zstatus = "updated%s." % forced
      else:
          zstatus = "unchanged."

      log.vv("%s %s" % (zfn, zstatus))

  xlat_slash_underscore = str.maketrans("/", "_")

  # Hosts header

  hostfile_html.append("<p><ul>")

  for net in list(Net.nets.values()):
      info = []
      if 'purpose' in net.attrs:
          info.append(net.attrs['purpose'])
      if 'addr' in net.__dict__:
          info.append(str(net.addr))

      hostfile_html.append("""<li><a href="#%s">%s</a> (%s)</li>
       """ % (net.name, net.name, ", ".join(info)))

  hostfile_html.append('</ul><small>')

  for net in list(Net.nets.values()):
      addr_usage = ''
      ipn4 = []

      total_addresses = 0
      free_addresses = 0
      for ipn, _ in sorted(net.ipnets,
                           key=lambda ipn__: ipn__[0].key()):
          hostfile.append('\n# %s - %s\n\n' % (net.name, ipn.prefix))
          hostfile.extend(ipn.list_hosts())

          if ipn.prefix.version == 4:
              total_addresses += ipn.prefix.numhosts
              free_addresses += len(ipn.freelist)
              ipn4.append(ipn)

      used_addresses = total_addresses - free_addresses
      if total_addresses > 0:
          addr_usage = "%d/%d available addresses" % \
                       (free_addresses, total_addresses)
      addr = " (%s)" % ', '.join(map(str, ipn4))
      hostfile_html.append("""
       <td><a name="%s"></a><h3>%s%s</h3>
       %s
       <table class="hosts">
      """ % (net.name, net.name, addr, addr_usage))
      if len(ipn4) > 0:
          for n in ipn4:
              hostfile_html.extend(n.html_list_hosts())
      else:
          for h in [h for h in list(Host.hosts.values()) if 'net' in h.attrs and h.attrs['net'] == net.name]:
              for a in h.addrs:
                  hostfile_html.append(h.html(a))

      hostfile_html.append("\n</table></td>")

  for net in list(Net.nets.values()):
      if len(net.dhcp_fixed) > 0:
          try:
              fixed_index = dhcpfile.index(
                  "# FIXED ASSIGNMENTS FOR %s\n" % (net.name))
              dhcpfile[fixed_index + 1:fixed_index + 1] = net.dhcp_fixed
          except ValueError:
              print("Could not find where to generate fixed DHCP assignments" \
                    + " for %s\n" \
                  % (net.name))
              raise

  gen_file('hosts', hostfile)
  gen_file('dhcpd.conf', dhcpfile)

  # 4. Generate HTML view of the machines

  # Create the HTML subdirectories if they do not exist

  for path in [html_subdir, html_machines_subdir]:
      if not os.path.isdir(path):
          os.mkdir(path)


  def html_line_for_host(host):
      """ Return a html table line or group of lines for host.
      """
      purpose = host.get_attr("purpose")
      if purpose == "":
          purpose = "(no description)"

      return """<tr><td><a href="machines/%s.html">%s</a></td>
                              <td>%s</td>
                         </tr>
          """ % (host.name, host.name, purpose)


  # Generate one file for each group
  for g in list(Group.groups.values()):
      groupdata = html_header(g.name) + """<h2>Group '%s'</h2> <h3>%s</h3>
                 """ % (g.name, g.get_attr("purpose"))

      # Write the hosts table
      groupdata += '<table class="hosts">'

      g.hosts.sort()
      for h in g.hosts:
          groupdata += html_line_for_host(h)

      groupdata += '</table>'

      gen_file(html_subdir + os.sep + g.name + ".html", groupdata)

  # Generate one file for all the machines that do not belong to any group

  groupdata = html_header("Ungrouped")
  groupdata += """<h2>Devices not associated to any group</h2>"""
  groupdata += '<table class="hosts">'

  for h in list(Host.hosts.values()):
      if h.groups == []:
          groupdata += html_line_for_host(h)

  groupdata += '</table>'
  gen_file(html_subdir + os.sep + "ungrouped.html", groupdata)

  # Generate one file for each host, with all the details

  for h in list(Host.hosts.values()):
      hostdata = html_header(h.name, subdir=1) + \
          "<h2>%s</h2>""" % h.name

      purpose = h.get_attr("purpose")
      if purpose != "":
          hostdata += """ Purpose:<ul><li>%s</li></ul>
          """ % purpose

      if h.groups != []:
          hostdata += """ Groups: <ul>%s</ul>
            """ % ("".join(["<li><a href=../%s.html>%s</a></li>"
                                         % (x, x) for x in h.groups]))

      if h.aliases != []:
          hostdata += """ Aliases: <ul>%s</ul>
            """ % ("".join(['<li>' + x[0] + ' (' + x[1] + ')</li>' for x in h.aliases]))

      if h.addrs != []:
          hostdata += """ Addresses: <ul>%s</ul>
             """ % ("".join(["<li>%s</li>" % (a) for a in h.addrs]))

      if h.mac_addrs != []:
          hostdata += """ MAC addresses: <ul>%s</ul>
           """ % ("<li> ".join(['<li><b>' + x[0] + '</b>: ' + x[1] + '</li>' for x in h.mac_addrs]))

      if h.attrs != {}:
          hostdata += """ Attributes: <ul> """
          for j in h.attrs:
              hostdata += "<li><b>%s</b>: %s</li>" % (j, h.attrs[j])
          hostdata += "</ul>"

      gen_file(html_machines_subdir + os.sep + h.name + ".html", hostdata)

  # Generate one table of all machines, ordered by IP address

  gen_file(html_subdir + os.sep + "hosts.html", hostfile_html)

  # Generate a table ordered by commission date

  dated = []    # contains tuples ('date', 'hostname')
  undated = []  # contains list of hostnames
  for h in list(Host.hosts.values()):
      if 'commissioned' in h.attrs:
          dated.append((h.attrs['commissioned'], h.name))
      else:
          if not ('role' in h.groups):
              undated.append(h.name)

  dated.sort()

  html = html_header("Hosts by date of commission") + """
    <h2>Dated hosts</h2>
    <table class="hosts">
    """

  for h in dated:
      html += """<tr>
       <td>%s</td><td><a href="machines/%s.html">%s</a></td>
       </tr>""" % (h[0], h[1], h[1])

  html += """</table><h2>Non-dated hosts</h2><table class="hosts">"""

  for h in undated:
      html += '<tr><td><a href="machines/%s.html">%s</a></td></tr>' % (h, h)

  gen_file(html_subdir + os.sep + "by_date.html", html)

  # Generate an index file

  index = html_header("Hosts") + """
     <h1>Machine inventory</h2>

     <h2>View by IP address</h2>
      <ul>
        <li> <a href="hosts.html">Hosts</a> </li>
      </ul>

     <h2>View by group</h2>
       <ul>
       <li> <a href="ungrouped.html">ungrouped</a>: not belonging to any group
       </li>

       """
  group_purpose = [(g.name, g.get_attr("purpose"))
                   for g in list(Group.groups.values())]
  group_purpose.sort()

  for g_name, g_purpose in group_purpose:
      index += '<li><a href="%s.html">%s</a>: %s</li>' % (
          g_name, g_name, g_purpose)

  index += """</ul>
     <h2>View by date of commission</h2>
      <ul>
        <li> <a href="by_date.html">Hosts by date</a> </li>
      </ul>

      <h2>Download CSVs</h2>
      <ul>
        <li> <a href="inventory.csv">Inventory</a> </li>
      </ul>

  """

  gen_file(html_subdir + os.sep + "index.html", index)

  # Generate a .csv file

  csv = ""

  # ... csv table headers

  csv += csv_sep.join(
      ["name", "commissioned"] +
      [k for k in Group.groups]) + '\n'

  # ... csv table body

  # Create a sublist of hosts relevant for listing in the inventory

  hosts = [h for h in list(Host.hosts.values())
           if 'role' not in h.groups and
           'virtual' not in h.groups and
           'commissioned' in h.attrs]

  for h in hosts:
      try:
          csv += h.name + csv_sep

          if 'commissioned' in h.attrs:
              csv += "%s%s" % (h.attrs['commissioned'], csv_sep)
          else:
              csv += csv_sep

          for g in Group.groups:
              if g in h.groups:
                  csv += '1' + csv_sep
              else:
                  csv += '0' + csv_sep
      except Exception:
          log.err("Failed to generate CSV for " +
                  h.name + str(h.attrs))

      csv += '\n'

  if not os.path.isdir(csv_subdir):
      os.mkdir(csv_subdir)

  gen_file(csv_subdir + os.sep + "inventory.csv", csv)


if __name__ == "__main__":
    main()
